"""
A class to keep track of "flowable" objects: substances, materials, products, or services which are exchanged between
activities or between an activity and a context.

The main point of this set is to manage CAS numbers and their associated synonyms. The CasNumber class accepts string
or numeric inputs and generates padded and non-padded synonyms that should match any reasonable CAS specification. It
also handles removing and serializing CAS numbers
"""


from .cas_number import CasNumber, cas_regex
from ..synonym_set import SynonymSet, DuplicateChild


class Flowable(SynonymSet):
    def add_term(self, term):
        """
        Any term that can be coerced into a CAS number is added AS a CAS number
        :param term:
        :return:
        """
        if bool(cas_regex.match(str(term))):
            self.add_child(CasNumber(term))
        else:
            super(Flowable, self).add_term(str(term))

    def remove_term(self, term):
        """
        If term is a CAS number, remove the whole child
        :param term:
        :return:
        """
        term = str(term)
        if bool(cas_regex.match(term)):
            try:
                self.remove_child(next(c for c in self.children if term in c))
            except StopIteration:
                pass
        else:
            super(Flowable, self).remove_term(str(term))

    def add_child(self, other, force=False):
        if isinstance(other, CasNumber):
            try:
                super(Flowable, self).add_child(other)  # duplicate CAS numbers also not allowed
            except DuplicateChild:
                return
        else:
            super(Flowable, self).add_child(other)

    @property
    def cas_numbers(self):
        for c in sorted(self._children, key=str):
            if isinstance(c, CasNumber):
                yield str(c)

    '''
    def set_name(self, name):
        raise NotSupported
    '''

    @property
    def object(self):
        return self

    def serialize(self):
        """
        Keep just names of child sets, since CAS numbers will auto-replicate on add, and other child sets should
        [normatively] behave similarly
        :return:
        """
        d = {
            'name': self._name,
            'synonyms': [t for t in sorted(self._terms) if t != self._name]
        }
        for c in sorted(self.children, key=lambda x:x.name):
            if c.name == self._name:
                continue
            d['synonyms'].append(c.name)
        return d



"""
It is unfortunate that "context managers" are a thing in python and other languages- but here they mean something
different.

So I'm going to bow to the reserved keyword and call these things compartments instead of contexts.


"""

from ..synonym_dict import SynonymDict
from .compartment import Compartment

NullCompartment = Compartment.null()

# these are not-really-subcompartments whose names should be modified if they have parents
NONSPECIFIC_LOWER = {'unspecified', 'non-specific', 'nonspecific', 'unknown', 'undefined', 'none'}


class NonSpecificCompartment(Exception):
    pass


class InconsistentLineage(Exception):
    pass


class CompartmentManager(SynonymDict):

    _entry_group = 'Compartments'
    _syn_type = Compartment
    _ignore_case = True

    _null_entry = NullCompartment

    def _add_from_dict(self, j):
        """
        JSON dict has mandatory 'name', optional 'parent', and 'synonyms'
        We POP from it because it only gets processed once
        :param j:
        :return:
        """
        name = j.pop('name')
        syns = j.pop('synonyms', [])
        parent = j.pop('parent', None)
        self.new_entry(name, *syns, parent=parent, **j)

    def load_dict(self, j):
        comps = j[self._entry_group]
        subs = []
        while len(comps) > 0:
            for obj in comps:
                if 'parent' in obj:
                    try:
                        self._d[obj['parent']]
                    except KeyError:
                        subs.append(obj)
                        continue
                self._add_from_dict(obj)
            comps = subs
            subs = []

    def _list_entries(self):
        return [c for c in self.objects]

    @property
    def top_level_compartments(self):
        for v in self.entries:
            if v.parent is None:
                yield v

    @property
    def objects(self):
        for tc in self.top_level_compartments:
            for c in tc.self_and_subcompartments:
                yield c

    def new_entry(self, *args, parent=None, **kwargs):
        """
        If a new object is added with unmodified non-specific synonyms like "unspecified", modify them to include their
        parents' name
        :param args:
        :param parent:
        :param kwargs:
        :return:
        """
        if parent is not None:
            if not isinstance(parent, Compartment):
                parent = self._d[parent]
            pn = parent.name
            args = tuple([', '.join([pn, k]) if k.lower() in NONSPECIFIC_LOWER else k for k in args])
        else:
            for k in args:
                if str(k).lower() in NONSPECIFIC_LOWER:
                    raise NonSpecificCompartment(k)
        return super(CompartmentManager, self).new_entry(*args, parent=parent, **kwargs)

    def _merge(self, existing_entry, ent):
        """
        Need to check lineage. We adopt the rule: merge is acceptable if both entries have the same top-level
        compartment or if ent has no parent.  existing entry will obviously be dominant
        :param existing_entry:
        :param ent:
        :return:
        """
        print('merging %s into %s' % (ent, existing_entry))
        if ent.parent is not None:
            if not ent.top() is existing_entry.top():
                raise InconsistentLineage('"%s": existing top %s | incoming top %s' % (ent,
                                                                                       existing_entry.top(),
                                                                                       ent.top()))

        super(CompartmentManager, self)._merge(existing_entry, ent)
        for sub in list(ent.subcompartments):
            sub.parent = existing_entry

    @staticmethod
    def _tuple_to_name(comps):
        return '; '.join(filter(None, comps))

    def _check_subcompartment_lineage(self, current, c):
        """
        Determines whether the incoming compartment name 'c' already exists in the database with an inconsistent
        lineage from the current parent 'current'.

        If the term is not found, creates a new subcompartment with current as parent.

        If the term is found and has a valid lineage, the found subcompartment is used.

        If the term is found to be an orphan, then current is assigned as its parent, unless the "orphan" is an
        elementary root context ('emissions' or 'resources'), in which case all terms in current + parents are removed
        from the dictionary and assigned to _disregarded

        Otherwise, raises InconsistentLineage
        :param current:
        :param c:
        :return:
        """
        if c in self._d:
            new = self.get(c)
            if current is None:
                return new
            if new.is_subcompartment(current):
                return new
            if new.parent is None:
                new.parent = current
                return new
            raise InconsistentLineage('"%s": existing parent "%s" | incoming parent "%s"' % (c,
                                                                                             new.parent,
                                                                                             current))
        else:
            new = self.new_entry(c, parent=current)
            return new

    def add_compartments(self, comps, conflict='rename'):
        """
        comps should be a list of Compartment objects or strings, in descending order
        :param comps:
        :param conflict: ['rename'] strategy to resolve inconsistent lineage problems.
          'rename' changes the name of the conflicting entry to include its native (nonconflicting) parent
          'match' hunts among the subcompartments of parent for a regex find
          'skip' simply drops the conflicting entry
          None or else: raise InconsistentLineage

        :return: the last (most specific) Compartment created
        """
        if len(comps) == 0 or comps is None:
            return self._null_entry
        current = None
        auto_name = self._tuple_to_name(comps)
        if auto_name in self._d:
            return self[auto_name]
        for c in comps:
            try:
                new = self._check_subcompartment_lineage(current, c)
            except InconsistentLineage as e:
                if conflict == 'match':
                    try:
                        new = next(s for s in current.subcompartments if s.contains_string(c, ignore_case=True))
                    except StopIteration:
                        raise e
                elif conflict == 'skip':
                    new = current
                elif conflict == 'rename':
                    new_c = ', '.join([current.name, c])
                    new = self._check_subcompartment_lineage(current, new_c)
                else:
                    raise e

            current = new
        self.add_synonym(current.name, auto_name)
        return current

    def __getitem__(self, item):
        if isinstance(item, tuple):
            item = self._tuple_to_name(item)
        if str(item).lower() in NONSPECIFIC_LOWER:
            return self._null_entry
        return super(CompartmentManager, self).__getitem__(item)



"""
A JSON-serializable collection of Flowable objects
"""

from .flowable import Flowable
from ..synonym_dict import SynonymDict


class FlowablesDict(SynonymDict):

    _entry_group = 'Flowables'
    _syn_type = Flowable
    _ignore_case = True

    def matching_flowables(self, *args):
        """
        just exposes the match_set function
        :param args:
        :return:
        """
        for k in self._match_set(args):
            yield k.name



"""
SynonymSet
A set of synonyms in a hashable collection
"""
import re
from uuid import uuid4


class RemoveNameError(Exception):
    pass


class ChildNotFound(Exception):
    pass


class DuplicateChild(Exception):
    pass


class SynonymSet(object):
    """

    """
    def __init__(self, *args, **kwargs):
        self._args = kwargs  # ignored
        self._terms = set()
        self._name = None
        self._children = set()

        self._id = str(uuid4())  # a private ID which exists only to hash the set consistently
        for arg in args:
            self.add_term(arg)

    @property
    def terms(self):
        if self.name is None:
            return
        seen = {self.name}
        yield self.name
        for s in self.base_terms:
            if s not in seen:
                yield s
                seen.add(s)
        for c in self._children:
            for t in c.terms:
                if t not in seen:
                    yield t
                    seen.add(t)

    @property
    def base_terms(self):
        for s in sorted(self._terms):
            yield s

    @property
    def children(self):
        for c in self._children:
            yield c

    @property
    def object(self):
        """
        The thing that the synonyms signify.  In the base class, it is just the name.  Create a Subclass to return
        objects of a different type.
        :return:
        """
        return self._name

    @property
    def name(self):
        return self._name

    def add_term(self, term):
        if hasattr(term, 'terms'):
            self.add_child(term)
        else:
            self._add_term(term)

    def _add_term(self, term):
        s = str(term)
        if len(s.strip()) == 0:
            return
        if self._name is None:
            s_s = s.strip()
            self._name = s_s
            self._terms.add(s_s)
        self._terms.add(s)

    def add_child(self, other, force=False):
        """
        This method stores the child as a subsidiary of the current object.  The child's terms will show up after the
        parent's terms.  Repeated terms will not be shown twice.  Child can still be modified.
        :param other:
        :param force: override restriction against having duplicate child sets (since child sets may still be
        individually updated
        :return:
        """
        if hasattr(other, 'terms'):
            if not force:
                if any(t == other for t in self._children) and len(other) > 0:
                    raise DuplicateChild('Child members match existing child')
            self._children.add(other)
            if self._name is None:
                self._name = other.name
        else:
            raise TypeError('Argument is not a synonym set (type %s)' % type(other))

    def remove_child(self, child):
        if child in self._children:
            self._children.remove(child)
        else:
            raise ChildNotFound('%s has no child %s' % (self, child))

    def children_with_term(self, term):
        """
        Returns all child sets having the given term
        :param term:
        :return: a tuple containing all child sets that contain the term
        """
        found = []
        for c in self._children:
            if term in c:
                found.append(c)
        return tuple(found)

    def contains_string(self, pattern, ignore_case=False):
        if ignore_case:
            srch = re.compile(pattern, flags=re.IGNORECASE)
        else:
            srch = re.compile(pattern)
        for t in self.terms:
            if bool(srch.search(t)):
                return t
        return False

    def has_child(self, other):
        return other in self._children

    def remove_term(self, term):
        s = str(term)
        _rmvd = False
        if s == self._name:
            raise RemoveNameError('%s is currently the name of this set.  Assign a different name before removing.' % s)
        if s in self._terms:
            self._terms.remove(s)
            _rmvd = True
        for c in self._children:
            if s in c:
                c.remove_term(s)
                _rmvd = True
        if _rmvd is False:
            raise KeyError('Unknown term %s' % term)

    def set_name(self, name):
        s = str(name)
        if s not in self._terms:
            self.add_term(s)  # is this desirable? Maybe we want to restrict this to names already in the set?
        self._name = s

    def serialize(self):
        """
        Note that this flattens child sets into self- allows for rewrites by assigning children
        :return:
        """
        j = {
            'name': self._name,
            'synonyms': [t for t in sorted(self.terms) if t != self._name]
        }
        if len(j['synonyms']) == 0:
            j.pop('synonyms')
        return j

    def __str__(self):
        return self._name

    def __repr__(self):
        return '%s(%s, %d terms, %d children)' % (self.__class__.__name__, self._name, len(self._terms),
                                                  len(self._children))

    def __len__(self):
        return len([t for t in self.terms])

    def __hash__(self):
        return hash(self._id)

    def __contains__(self, item):
        return item in self.terms

    def __eq__(self, other):
        if hasattr(other, 'terms'):
            otherterms = set(other.terms)
        else:
            otherterms = {other}
        return set(self.terms) >= otherterms  # yes, I know, that's not equal per se



from ..synonym_dict import SynonymDict, TermExists, MergeError
from ..synonym_set import SynonymSet, RemoveNameError
import unittest


class TestContainer(object):
    """
    This gimmick allows the test to be inherited without getting discovered twice by unittest
    """
    class SynonymDictTest(unittest.TestCase):

        _test_class = SynonymDict
        # include explicit 'ignore_case=False' for tests that depend on it so they can be skipped in subclasses

        def test_new_object(self):
            g = self._test_class()
            g.new_entry('hello', 'hola', 'hi')
            self.assertEqual(g.get('hola'), 'hello')

        def test_add_object(self):
            g = self._test_class()
            g.new_entry('hello', 'hola', 'hi')
            s = SynonymSet('goodbye', 'au revoir', 'adios')
            g.add_or_update_entry(s)
            self.assertEqual(g.get('adios'), 'goodbye')

        def test_implicit_merge(self):
            g = self._test_class()
            g.new_entry('hello', 'hola', 'hi', 'aloha')
            with self.assertRaises(TermExists):
                g.new_entry('goodbye', 'au revoir', 'aloha', merge=False)
            g.new_entry('goodbye', 'au revoir', 'adios')
            with self.assertRaises(MergeError):
                g.new_entry('yes', 'no', 'goodbye', 'hello')
            g.new_entry('hi', 'Ni hao', 'hallo')
            self.assertEqual(g['hello'], g['hallo'])

        def test_explicit_merge(self):
            g = self._test_class(ignore_case=False)
            o = g.new_entry('hello', 'hola', 'hi', 'aloha')
            p = g.new_entry('Hello', 'HELLO', 'Hi', 'HI')
            self.assertSetEqual({k for k in g.objects}, {o.object, p.object})
            self.assertNotEqual(g['hi'], g['HI'])
            g.merge('hi', 'HI')
            self.assertEqual(g['hi'], g['HI'])
            self.assertListEqual([k for k in g.objects], [o.object])

        def test_child(self):
            g = self._test_class()
            ob1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            ob2 = g.new_entry('hi', 'Ni hao', 'hallo', create_child=True)
            self.assertTrue(ob1.has_child(ob2))
            ob3 = g.new_entry('greetings', 'salutations', 'hello')
            self.assertFalse(ob1.has_child(ob3))
            self.assertEqual(g.get('greetings'), g.get('hola'))

        def test_objects(self):
            g = self._test_class(ignore_case=False)
            o1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            o2 = g.new_entry('Hello', 'HELLO', 'Hi', 'HI')
            self.assertListEqual([k for k in g.objects], [o2, o1])  # sort order by string
            #self.assertListEqual([k for k in g.objects], [o2.object, o1.object])  # sort order by string

        def test_remove_child(self):
            g = self._test_class()
            ob1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            ob2 = g.new_entry('hi', 'Ni hao', 'hallo', create_child=True)
            self.assertEqual(g['hello'], g['hallo'])
            with self.assertRaises(TermExists):
                g.unmerge_child(ob2)
            self.assertTrue(ob1.has_child(ob2))
            with self.assertRaises(RemoveNameError):
                ob2.remove_term('hi')
            ob2.set_name('hallo')
            ob2.remove_term('hi')
            g.unmerge_child(ob2)
            self.assertNotEqual(g['hello'], g['hallo'])
            self.assertEqual(g['Ni hao'], str(ob2))

        def test_case_insensitive(self):
            g = self._test_class(ignore_case=True)
            g.new_entry('hello', 'hola', 'hi', 'aloha')
            g.new_entry('Hi', 'Bonjour')
            self.assertEqual(g['bonjour'], 'hello')

        def test_remove_term_case_insensitive(self):
            g = self._test_class(ignore_case=True)
            o = g.new_entry('hello', 'hola', 'hi', 'aloha')
            self.assertIn('Hi', g)
            self.assertIn('hi', o)
            self.assertNotIn('Hi', o)
            g.del_term('Hi')
            self.assertNotIn('Hi', g)
            self.assertNotIn('hi', o)

        def test_whitespace(self):
            g = self._test_class(ignore_case=True)
            g.new_entry('hello', 'bonjour')
            self.assertEqual(g[' bonjour'], 'hello')
            g.add_synonym('bonjour ', ' GREETINGS  ')
            self.assertEqual(g['greetings'], 'hello')

        def test_disallow_empty_string_silently(self):
            g = self._test_class()
            g.new_entry('hello', 'bonjour')
            g.add_synonym('hello', '')
            self.assertNotIn('', g)

        def test_serialize(self):
            g = self._test_class()
            g.new_entry('hello', 'bonjour')
            g.new_entry('adios', 'adieu')
            g.serialize()

        def test_cast_to_str(self):
            g = self._test_class(ignore_case=False)
            a = g.new_entry('the answer', 42)
            self.assertEqual(g['42'], 'the answer')
            self.assertIsNone(g.get(42))
            self.assertIn('42', a)
            self.assertNotIn(42, a)

        def test_remove_object(self):
            g = self._test_class(ignore_case=False)
            o1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            o2 = g.new_entry('Hello', 'HELLO', 'Hi', 'HI')
            self.assertEqual(g['HELLO'], str(o2))
            g.remove_entry(o2)
            self.assertListEqual([k for k in g.objects], [o1])
            with self.assertRaises(KeyError):
                _ = g['HELLO']

        def test_synonyms(self):
            g = self._test_class(ignore_case=True)
            o1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            o2 = g.new_entry('Hello', 'HELLO', 'Hi', 'HI')
            self.assertIs(o1, o2)
            self.assertSetEqual(set(g.synonyms('hello')), set(o1.terms))
            self.assertSetEqual(set(g.synonyms(o2)), set(o1.terms))

        def test_add_synonym(self):
            g = self._test_class()
            o1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            g.add_synonym('hi', 'greetings')
            self.assertIn('greetings', o1)
            self.assertEqual(g['greetings'], g['hello'])

        def test_add_synonym_entry(self):
            g = self._test_class()
            o1 = g.new_entry('hello', 'hola')
            self.assertIsInstance(o1, g._syn_type)
            g.add_synonym(o1, 'greetings')
            self.assertEqual(g['greetings'], g['hola'])

        def test_prune(self):
            g = self._test_class()
            o1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            o2 = g.new_entry('goodbye', 'farewell', 'aloha', prune=True)
            self.assertIn('goodbye', g)
            self.assertEqual(g['aloha'], 'hello')
            self.assertEqual(g['farewell'], 'goodbye')
            self.assertIn('aloha', o1)
            self.assertNotIn('aloha', o2)

        def test_prune_idempotent(self):
            g = self._test_class()
            o1 = g.new_entry('hello', 'hola', 'hi', 'aloha')
            o2 = g.new_entry('hello', 'hola', 'hi', 'aloha', prune=True)
            self.assertIs(o2, o1)

        def test_pruned_set_name_closure(self):
            g = self._test_class()
            g.new_entry('hello', 'hola', 'hi', 'aloha')
            g.new_entry('goodbye', 'adios')
            with self.assertRaises(MergeError):
                g.new_entry('aloha', 'goodbye', 'farewell', 'sayonara')
            o3 = g.new_entry('aloha', 'goodbye', 'farewell', 'sayonara', prune=True)
            self.assertIn(o3.name, g.synonyms(o3))

        def test_type_consistency(self):
            """
            new_entry should return entry, get and __getitem__ should return object
            :return:
            """
            g = self._test_class()
            o1 = g.new_entry('a', 'b', 'c')
            o2 = g['a']
            o3 = g.get('b')
            self.assertEqual(o1.object, o2)
            self.assertEqual(o1.object, o3)
            self.assertEqual(type(o1.object), type(o2))
            self.assertEqual(type(o2), type(o3))

class SynonymDictTest(TestContainer.SynonymDictTest):
    pass


if __name__ == '__main__':
    unittest.main()



from .cas_number import CasNumber, InvalidCasNumber, NotSupported
import unittest


co2_set = {'000124-38-9', '000124389', '124-38-9', '124389'}


class CasNumberTest(unittest.TestCase):
    def test_valid_co2(self):
        g = CasNumber(124389)
        self.assertSetEqual(set([k for k in g.terms]), co2_set)
        self.assertEqual(str(g), '124-38-9')

    def test_invalid(self):
        with self.assertRaises(InvalidCasNumber):
            CasNumber(1234)
        with self.assertRaises(InvalidCasNumber):
            CasNumber(55, 89, 47, 2)
        with self.assertRaises(InvalidCasNumber):
            CasNumber('37205.543.1')

    def test_add_remove(self):
        c = CasNumber(12345)
        with self.assertRaises(NotSupported):
            c.add_term('Floobie')
        self.assertIn('12-34-5', c)
        with self.assertRaises(NotSupported):
            c.remove_term('12-34-5')

    def test_equal(self):
        g = CasNumber(124389)
        h = CasNumber('124-38-9')
        self.assertEqual(g, h)

    def test_pad_input(self):
        g = CasNumber('00124+38+9')
        self.assertSetEqual(set([k for k in g.terms]), co2_set)

    def test_tuple_input(self):
        g = CasNumber(('32768', '4', 1))
        self.assertEqual(str(g), '32768-04-1')
        h = CasNumber((32768, 4, '1'))
        self.assertEqual(h.object, g.object)
        j = CasNumber('32768', 4.0, '1')
        self.assertEqual(h.object, j.object)


if __name__ == '__main__':
    unittest.main()



from synonym_dict.tests.test_synonym_dict import TestContainer
from .compartment import Compartment
from .compartment_manager import CompartmentManager, NonSpecificCompartment, InconsistentLineage
import unittest


class CompartmentContainer(object):
    class CompartmentTest(unittest.TestCase):

        def test_parent(self):
            c = Compartment('emissions')
            d = Compartment('emissions to air', parent=c)
            e = Compartment('emissions to urban air', parent=d)
            f = Compartment('emissions to rural air', 'emissions from high stacks', parent=d)
            self.assertSetEqual(set(i for i in d.subcompartments), {e, f})
            self.assertListEqual([str(k) for k in c.self_and_subcompartments], ['emissions', 'emissions to air',
                                                                                'emissions to rural air',
                                                                                'emissions to urban air'])

        def test_serialize(self):
            c = Compartment('emissions')
            d = Compartment('emissions to air', parent=c)
            j = d.serialize()
            self.assertEqual(j['name'], 'emissions to air')
            self.assertSetEqual(set(j.keys()), {'name', 'synonyms', 'parent'})

        def test_iter(self):
            c = Compartment('emissions')
            d = Compartment('emissions to air', parent=c)
            e = Compartment('emissions to urban air', parent=d)
            self.assertListEqual(e.as_list(), ['emissions', 'emissions to air', 'emissions to urban air'])
            self.assertTupleEqual(tuple(e), ('emissions', 'emissions to air', 'emissions to urban air'))


    class CompartmentManagerTest(TestContainer.SynonymDictTest):
        """
        What do we want compartment managers to do?

         - one, keep track of compartments that are encountered
          = add them (hierarchicaly)
          = add synonyms as appropriate
         - two, retrieve a compartment from a string (the syndict already handles this)
         - three, that's really it.  managing a set of canonical compartments is a separate task.
        """
        def _test_class(self, ignore_case=True):
            if ignore_case is False:
                self.skipTest('skipping case sensitive test')
            else:
                return CompartmentManager()

        def test_add_object(self):
            with self.assertRaises(TypeError):
                super(CompartmentContainer.CompartmentManagerTest, self).test_add_object()

        def setUp(self):
            self.cm = CompartmentManager()
            #

        def _add_water_dict(self):
            self.cm.new_entry('Emissions')
            d = {'name': 'water emissions',
                 'synonyms': [
                     'emissions to water',
                     'emissions to surface water',
                     'water'
                 ],
                 'parent': 'emissions'}
            self.cm._add_from_dict(d)

        def test_add_from_dict(self):
            self._add_water_dict()
            self.assertEqual(str(self.cm['water']), 'water emissions')

        def test_0_add_hier(self):
            self.cm.add_compartments(['emissions', 'emissions to air', 'emissions to urban air'])
            self.assertIs(self.cm['emissions to air'], self.cm['emissions to urban air'].parent)

        def test_strip_name(self):
            c = self.cm.add_compartments(['NETL Elementary Flows', ' [Resources] '])
            self.assertEqual(c.name, '[Resources]')

        def test_idempotent(self):
            """
            getting a context should return the context
            :return:
            """
            ca = self.cm.new_entry('Resources')
            cx = self.cm['resources']
            self.assertIs(self.cm[cx], cx, ca)

        def test_null(self):
            cx = self.cm[None]
            self.assertIs(cx, self.cm._null_entry)

        def test_add_null(self):
            cx = self.cm.add_compartments(())
            self.assertIs(cx, self.cm._null_entry)

        def test_toplevel(self):
            self.cm.add_compartments(['social hotspots', 'labor', 'child labor'])
            self.assertIn('social hotspots', (str(x) for x in self.cm.top_level_compartments))

        def test_unspecified(self):
            c = self.cm.add_compartments(['emissions', 'water', 'unspecified'])
            self.assertEqual(c.name, 'water, unspecified')
            self.assertEqual(c.parent.name, 'water')

        def test_top_level_nonspecific(self):
            with self.assertRaises(NonSpecificCompartment):
                self.cm.add_compartments(['unspecified', 'unspecified water'])

        def test_retrieve_by_tuple(self):
            self._add_water_dict()
            w = self.cm['water']
            self.assertIs(w, self.cm[('emissions', 'water emissions')])

        '''
        Potential Glitch cases:
         * relative add
         * omitted descendant -> still valid
         * conflict in specified parent -> InconsistentLineage
        '''
        def test_relative_add(self):
            self._add_water_dict()
            uw = self.cm['water']
            ud = self.cm.add_compartments(['water', 'lake water'])
            self.assertIs(uw, ud.parent)
            self.assertListEqual(ud.as_list(), ['Emissions', 'water emissions', 'lake water'])

        def test_omitted_descendant(self):
            ua = self.cm.add_compartments(['emissions', 'to air', 'to urban air'])  # confirm that this exists
            uc = self.cm.add_compartments(['emissions', 'to urban air', 'to urban center'])
            self.assertIs(ua, uc.parent)

        def test_inconsistent_lineage(self):
            self._add_water_dict()
            with self.assertRaises(InconsistentLineage):
                self.cm.add_compartments(['resources', 'water'], conflict=None)

        def test_inconsistent_lineage_match(self):
            """
            When an intermediate descendant conflicts, we can either raise the exception (cautious) or do some clever
            regex-based predictive guessing (reckless)
            :return:
            """
            self._add_water_dict()
            rw = self.cm.add_compartments(['resources', 'from water'])
            fw = self.cm.add_compartments(['resources', 'water', 'fresh water'], conflict='match')

            self.assertIs(fw.parent, rw)

        def test_inconsistent_lineage_skip(self):
            """
            When an intermediate descendant conflicts, we can either raise the exception (cautious) or do some clever
            regex-based predictive guessing (reckless)
            :return:
            """
            self._add_water_dict()
            rw = self.cm.add_compartments(['resources', 'from water'])
            fw = self.cm.add_compartments(['resources', 'water', 'fresh water'], conflict='skip')

            self.assertIs(fw.parent, rw.parent)

        def test_inconsistent_lineage_rename(self):
            """
            If we get an inconsistent lineage error on a subcompartment, this is caused by a synonym collision. We can solve
            it by renaming the new subcompartment to be based on its parent, which must be non-conflicting.

            For rename we use a comma-separated join so that semicolon remains the operative parent-child separator
            :return:
            """
            self._add_water_dict()
            self.cm.new_entry('Resources')
            rw = self.cm.add_compartments(['resources', 'water'], conflict='rename')
            self.assertIs(rw.parent, self.cm['resources'])
            self.assertEqual(rw.name, 'Resources, water')


class CompartmentTest(CompartmentContainer.CompartmentTest):
    pass


class CompartmentManagerTest(CompartmentContainer.CompartmentManagerTest):
    pass


if __name__ == '__main__':
    unittest.main()



"""
accept a variety of inputs; store synonyms for a CAS number of the canonical form xxxxxx-xx-x
"""

import re
from ..synonym_set import SynonymSet


class NotSupported(Exception):
    pass


cas_regex = re.compile('^0*([0-9]{2,6})(\W?)([0-9]{2})\\2([0-9])$')


class InvalidCasNumber(Exception):
    """
    raised when input cannot be deciphered as a valid CasNumber
    """
    pass


'''
def _validate_numeric_input(cas):
    if cas < 10**4 or cas >= 10**9:
        raise InvalidCasNumber('Numeric input out of range')
    tup2 = cas % 10
    cas = int((cas - tup2) / 10)

    tup1 = cas % 100

    tup0 = int((cas - tup1) / 100)
    return str(tup0), str(tup1), str(tup2)
'''

def _validate_string_input(cas):
    match = cas_regex.match(cas)
    if match is None:
        raise InvalidCasNumber('String does not match regex')
    groups = match.groups()  # added a new group; need to filter it out
    return tuple(groups[k] for k in (0, 2, 3))


def _validate_tuple_input(tup):
    x = [int(k) for k in tup]
    y = '%d' % x[0], '%02d' % x[1], '%d' % x[2]
    if not bool(cas_regex.match('-'.join(y))):
        raise InvalidCasNumber('Tuple input does not match regex')
    return y


def _generate_cas_formats(tup):
    pad_tup = ('%06d' % int(float(tup[0])), tup[1], tup[2])
    for k in (tup, pad_tup):
        yield '-'.join(k)
        yield ''.join(k)


class CasNumber(SynonymSet):
    def __init__(self, *args):
        """
        Accept the following forms:
         - integer satisfying 10**4 <= x <= 10**8 (124389 = 124-38-9 Carbon Dioxide)
         - string matching the cas_regex '^0*([0-9]{2,6})\W?([0-9]{2})\W?([0-9])$'
         - 3-tuple containing either number or string having (6 or fewer digits, 2 digits, 1 digit)
         - 3 positional arguments that make up a 3-tuple according to above

        Produces a synonym set containing both zero-padded and non-zero-padded CAS number strings, both with and without
        hyphens.  e.g. above input would yield {'124-38-9', '124389', '000124-38-9', '000124389'}
        :param cas:
        """
        if len(args) == 1:
            cas = args[0]
        else:
            cas = tuple(args)

        if isinstance(cas, tuple):
            _tup = _validate_tuple_input(cas)
        else:
            _tup = _validate_string_input(str(cas))

        self._init = True

        super(CasNumber, self).__init__(*_generate_cas_formats(_tup))
        self._init = False

    def add_term(self, term):
        if self._init:
            super(CasNumber, self).add_term(term)
        else:
            raise NotSupported('May not add terms to CAS numbers')

    def remove_term(self, term):
        raise NotSupported('May not remove terms from CAS numbers')

    def add_child(self, other, force=False):
        raise NotSupported('CAS numbers may not have children')

    def set_name(self, name):
        raise NotSupported('May not change canonical name for CAS numbers')



from .lower_dict import LowerDict
from .synonym_set import SynonymSet, ChildNotFound

from collections import defaultdict
import json


class TermExists(Exception):
    """
    Thrown when adding a term that is already in the synonym set
    """
    pass


class MergeError(Exception):
    """
    Thrown when a merge would result in merging two separate sets
    """
    pass


class ParentNotFound(Exception):
    pass


class SynonymDict(object):
    """
    A class that allows retrieval of a given object by any of its synonyms.
    """
    _entry_group = 'SynonymSets'
    _syn_type = SynonymSet
    _ignore_case = False

    def _add_from_dict(self, j):
        name = j['name']
        syns = j.pop('synonyms', [])
        self.new_entry(name, *syns, merge=True)

    def load_dict(self, j):
        for ent in j[self._entry_group]:
            self._add_from_dict(ent)

    def load(self, filename=None):
        if filename is None:
            if self._filename is None:
                return
            filename = self._filename
        with open(filename, 'r') as fp:
            fbs = json.load(fp)
        self.load_dict(fbs)

    def _list_entries(self):
        return [f for f in self.entries]

    def serialize(self, entries=None):
        if entries is None:
            entries = self._list_entries()
        return {self._entry_group: [f.serialize() for f in entries]}

    def save(self, filename=None):
        """
        if filename is specified, it overrules any prior filename
        :param filename:
        :return:
        """
        if filename is not None:
            self._filename = filename
        with open(self._filename, 'w') as fp:
            json.dump(self.serialize(), fp, indent=2)

    @property
    def entry_group(self):
        return self._entry_group

    @entry_group.setter
    def entry_group(self, value):
        self._entry_group = str(value)

    def __init__(self, source_file=None, ignore_case=None, entry_group=None):
        """
        Create a synonym dictionary that stores entries of a certain type.
        :param ignore_case: [False] whether the synonyms should be considered case-sensitive. This setting also causes
         the dict to ignore leading and trailing whitespace (calls strip().lower())

        The entry class must support the following API:
          constructor:
            Ent(*args, **kwargs): kwargs not used for the base SynonymSet
          properties:
            ent.terms: generate a list of synonyms
            ent.object: return the payload object (for SynonymSet, this is a str)
          methods:
            ent.add_term(term): add a new synonym
            ent.remove_term(term): remove an existing synonym
            ent.set_name(term): assign the canonical name for the object (by default the first term provided)
            ent.add_child(a): include a's terms with ent's terms
        """
        self._filename = source_file
        if entry_group is not None:
            self.entry_group = str(entry_group)

        ignore_case = ignore_case or self._ignore_case

        if ignore_case:
            self._d = LowerDict()
            self._l = defaultdict(LowerDict)  # reverse mapping
        else:
            self._d = dict()
            self._l = defaultdict(dict)  # reverse mapping

        self.load()

    def _throw_term_exists(self, term):
        raise TermExists('%s -> %s' % (term, self._d[term]))

    def _check_term(self, term, check_obj=None):
        """
        If term is not known to the dict, returns None
        if check_obj is provided and term does NOT map to check_obj, throw a TermExists error
        if otherwise, return the entry that term maps to
        :param term:
        :param check_obj:
        :return:
        """
        if term in self._d:
            if check_obj is not None:
                if self._d[term] is not check_obj:
                    self._throw_term_exists(term)
            return self._d[term]

    def _add_term(self, term, ent):
        if not isinstance(ent, self._syn_type):
            raise TypeError('Entry is not a %s (%s)' % (self._syn_type, type(ent)))
        if len(term.strip()) == 0:
            return
        self._check_term(term, ent)
        self._d[term] = ent
        self._l[ent][term] = term

    def _remove_term(self, term, remove_from_entry=False):
        ent = self._d.pop(term)
        cterm = self._l[ent].pop(term)
        if remove_from_entry:
            ent.remove_term(cterm)
        if len(self._l[ent]) == 0:
            self._l.pop(ent)

    @property
    def entries(self):
        for k in sorted(self._l.keys(), key=str):
            yield k

    @property
    def objects(self):
        """
        Yield all distinct objects in the dict
        :return:
        """
        for k in self.entries:
            yield k.object

    def objects_with_string(self, pattern):
        for ent in self.entries:
            if ent.contains_string(pattern, ignore_case=self._ignore_case):
                yield ent.object

    def new_entry(self, *args, merge=True, create_child=False, prune=False, **kwargs):
        """
        The input arguments are passed directly to the constructor
        :param args: args to pass to constructor
        :param merge: [True] if synonyms are found in an existing entry, merge the new terms according to create_child
        :param create_child: [True] if true, add new entry as a child; otherwise, add terms to the existing entry
        :param prune: [False] If merge is False and a collision exists, simply omit conflicting terms from the new
         entry. Specifying prune=True to new_entry will override merge=True.
        :param kwargs: kwargs to pass to constructor
        :return: the entry that contains the new terms
        """
        ent = self._syn_type(*args, **kwargs)
        if prune:
            merge = False
        return self.add_or_update_entry(ent, merge=merge, create_child=create_child, prune=prune)

    def _match_set(self, terms):
        """
        Returns a set of synonym entries that match terms in the input argument.
        :param terms: iterable
        :return:
        """
        return set(k for k in filter(None, (self._check_term(t) for t in terms)))

    def match_entry(self, *args):
        mg = self._match_set(args)
        if len(mg) > 1:
            raise MergeError('Found terms in multiple sets: %s' % '; '.join(str(k) for k in mg))
        elif len(mg) == 1:
            return mg.pop()
        else:
            return None

    def add_or_update_entry(self, ent, merge=True, create_child=False, prune=False):
        """
        Returns the entry that contains the input argument's terms
        If merge is true, merge incoming entries with existing matching entry. If create_child, the new entry is kept
        intact and added as a child. If more than one existing entry matches, raise MergeError.

        If prune, existing terms are skipped and new terms are added to a new entry.

        merge and prune are mutually inconsistent.  merge wins.
        :param ent:
        :param merge:
        :param create_child:
        :param prune:
        :return: the entry that contains all of ent's terms
        """
        if merge:
            s = self.match_entry(*ent.terms)
            if s is None:
                for t in ent.terms:
                    self._add_term(t, ent)
                return ent
            else:
                if create_child:
                    self._add_child(s, ent)
                    return ent
                else:
                    self._merge(s, ent)
                    return s
        elif prune:
            prune_terms = set()
            for t in ent.terms:
                try:
                    self._check_term(t, ent)
                except TermExists:
                    prune_terms.add(t)
            if ent.name in prune_terms:
                try:
                    ent.set_name(next(k for k in ent.terms if k not in prune_terms))
                except StopIteration:
                    # no non-pruned terms: nothing to add
                    return self._d[ent.name]  # give back the current match
            for k in prune_terms:
                ent.remove_term(k)
            return self.add_or_update_entry(ent, merge=False, prune=False)
        else:
            for t in ent.terms:
                self._check_term(t, ent)
            for t in ent.terms:
                self._add_term(t, ent)
            return ent

    def _add_child(self, existing_entry, child):
        if existing_entry not in self._l.keys():
            raise ParentNotFound('Entry %s not known' % existing_entry)
        existing_entry.add_child(child)
        for t in child.terms:
            self._add_term(t, existing_entry)

    def _merge(self, existing_entry, ent):
        for c in ent.children:
            self._add_child(existing_entry, c)
        for t in ent.base_terms:
            if self._check_term(t) is None:
                self._add_term(t, existing_entry)
                existing_entry.add_term(t)

    def remove_entry(self, ent):
        """
        remove an entry and all of its terms from the dictionary, without altering the entry itself
        :param ent:
        :return:
        """
        remove_terms = [t for t in self._l[ent].keys()]
        for t in remove_terms:
            self._remove_term(t)

    def merge(self, first, second, child=False):
        """
        Merge the sets containing the two terms
        :param first: dominant
        :param second: removed from dict and merged with first
        :param child: [False] if true, preserve the second entry as a child, rather than merging its content
        :return:
        """
        ob1 = self._d[first]
        ob2 = self._d[second]
        self.remove_entry(ob2)
        if child:
            self._add_child(ob1, ob2)
        else:
            self._merge(ob1, ob2)

    def unmerge_child(self, ent):
        """
        The supplied ent must be a child of another entry.  Removes it and establishes it as a separate entry.  This
        will fail [TermExists] if the child and the parent share any terms in common.
        :param ent:
        :return:
        """
        try:
            parent = next(v for v in self.entries if v.has_child(ent))
        except StopIteration:
            raise ChildNotFound('No set found containing child %s' % ent)
        parent.remove_child(ent)
        dups = [k for k in ent.terms if k in parent]
        if len(dups) > 0:
            parent.add_child(ent)
            raise TermExists('Terms duplicated in parent: %s' % '; '.join(dups))
        for k in ent.terms:
            self._remove_term(k)
        return self.add_or_update_entry(ent, merge=False)

    def add_synonym(self, term, syn):
        """
        Add a new term as a synonym to an existing term
        :param term: the existing term
        :param syn: the new synonym
        :return:
        """
        syn = str(syn)
        if isinstance(term, self._syn_type):
            term = term.name
        ent = self._d[term]
        self._add_term(syn, ent)  # checks TermExists
        ent.add_term(syn)

    def del_term(self, term):
        """
        Delete a term from the dictionary and from the entry itself
        :param term:
        :return:
        """
        self._remove_term(term, remove_from_entry=True)

    def set_name(self, term):
        if term not in self:
            raise KeyError('Unknown term %s' % term)
        ent = self._d[term]
        cterm = self._l[ent][term]  # necessary for case-insensitive dictionaries
        ent.set_name(cterm)

    def get(self, term, default=None):
        try:
            return self.__getitem__(term)
        except KeyError:
            return default

    def synonyms(self, term):
        if isinstance(term, SynonymSet):
            ent = term
        else:
            ent = self._d[term]
        for t in sorted(self._l[ent].values()):
            yield t

    def items(self):
        for k, v in self._d.items():
            yield k, v.object

    def __len__(self):
        return len(self._l)

    def __contains__(self, item):
        return item in self._d

    def __getitem__(self, term):
        if isinstance(term, self._syn_type) and term in self._l:
            return term.object
        return self._d[term].object



from ..synonym_set import SynonymSet, RemoveNameError, DuplicateChild
import unittest


class SynonymSetTest(unittest.TestCase):
    def test_empty(self):
        s = SynonymSet()
        self.assertListEqual([k for k in s.terms], [])

    def test_add_empty(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        self.assertNotIn('', s)
        s.add_term('')
        self.assertNotIn('', s)

    def test_terms(self):
        """
        first term is name, always shows first
        remaining terms show in alphabetical order
        :return:
        """
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        self.assertListEqual(['hello', 'Ni hao', 'aloha'], [k for k in s.terms])

    def test_name(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        self.assertEqual(str(s), 'hello')
        self.assertEqual(s.object, 'hello')
        s.set_name('aloha')
        self.assertEqual(s.object, 'aloha')

    def test_add(self):
        s = SynonymSet('the answer', 'the answer to the ultimate question', 42)
        s.add_term('the ultimate answer')
        self.assertEqual(len([k for k in s.terms]), 4)

    def test_remove(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        with self.assertRaises(RemoveNameError):
            s.remove_term('hello')
        s.set_name('Ni hao')
        s.remove_term('hello')
        self.assertEqual(len([k for k in s.terms]), 2)

    def test_child(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        t = SynonymSet('bonjour', 'hola', 'hello', 'Hi')
        s.add_term(t)
        terms = [k for k in s.terms]
        self.assertTrue(s.has_child(t))
        self.assertEqual(len(terms), 6)
        self.assertSetEqual(set(terms), {'hello', 'aloha', 'Ni hao', 'bonjour', 'hola', 'Hi'})

    def test_remove_from_child(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        t = SynonymSet('bonjour', 'hola', 'hello', 'Hi')
        s.add_child(t)
        s.remove_term('hola')
        self.assertNotIn('hola', t)
        self.assertNotIn('hola', s)

    def test_contains(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        self.assertTrue('aloha' in s)
        self.assertFalse('greetings' in s)

    def test_strip(self):
        s = SynonymSet(' [Resources] ')
        self.assertEqual(s.name, '[Resources]')
        self.assertIn(' [Resources] ', s)

    def test_contains_string(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        self.assertEqual(s.contains_string('hao'), 'Ni hao')
        self.assertFalse(s.contains_string('goober'))

    def test_hashable(self):
        s = SynonymSet('hello', 'aloha', 'Ni hao')
        t = SynonymSet('bonjour', 'hola', 'hello', 'Hi')
        g = {s, t}
        self.assertEqual(len(g), 2)

    def test_avoid_duplicate_subsets(self):
        s = SynonymSet('hello', 'hi', 'greetings', 'salutations')
        t = SynonymSet('bonjour', 'aloha', 'ni hao', 'hola')
        u = SynonymSet('hola', 'bonjour', 'ni hao', 'aloha')
        s.add_term(t)
        self.assertEqual(len(s._children), 1)
        with self.assertRaises(DuplicateChild):
            s.add_term(u)
        self.assertEqual(len(s._children), 1)
        self.assertFalse(s.has_child(u))


if __name__ == '__main__':
    unittest.main()



"""
Case-insensitive dict
from https://stackoverflow.com/questions/2082152
(see also https://stackoverflow.com/questions/3387691 -- my requirements are specific enough to justify this approach,
so long as it works)
"""


class LowerDictKeys(object):
    def __init__(self, dict_keys):
        self._orig = dict_keys
        self._dk = [k.lower() for k in dict_keys]

    def __contains__(self, item):
        return item.strip().lower() in self._dk

    def __iter__(self):
        for k in self._orig:
            yield str(k)


class LowerDict(dict):
    """
    A case-insensitive dict.  Works just like a regular dict, except key hash and equality are
    tested against the strip().lower() version of the key.

    This is nice because the dict still remembers the authentic key, with capitalization as-entered,
    and can return it when asked.
    """

    class Key(str):
        def __new__(cls, key):
            # perform the strip() transform at instantiation for performance
            return super(LowerDict.Key, cls).__new__(LowerDict.Key, str(key).strip())

        def __hash__(self):
            return hash(self.lower())

        def __eq__(self, other):
            return self.lower() == other.strip().lower()

    def __init__(self, *args, **kwargs):
        super(LowerDict, self).__init__()
        if len(args) > 0:
            for k, v in args[0]:
                self[k] = v
        for key, val in kwargs.items():
            self[key] = val

    def __contains__(self, key):
        key = self.Key(key)
        return super(LowerDict, self).__contains__(key)

    def __setitem__(self, key, value):
        key = self.Key(key)
        super(LowerDict, self).__setitem__(key, value)

    def __getitem__(self, key):
        key = self.Key(key)
        return super(LowerDict, self).__getitem__(key)

    def keys(self):
        return LowerDictKeys(super(LowerDict, self).keys())

    def items(self):
        for k, v in super(LowerDict, self).items():
            yield str(k), v

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def pop(self, key, default=None):
        key = self.Key(key)
        return super(LowerDict, self).pop(key, default)

    def update(self, m, **kwargs):
        for k, v in m.items():
            self.__setitem__(k, v)
        for k, v in kwargs.items():
            self.__setitem__(k, v)



from ..lower_dict import LowerDict
import unittest


class LcEntityTest(unittest.TestCase):
    def test_iterable(self):
        e = LowerDict((('a', 1), ('b', 2), ('c', 3)))
        self.assertEqual(e['A'], 1)
        self.assertEqual(e['C'], 3)
        with self.assertRaises(KeyError):
            print(e['d'])

    def test_get(self):
        e = LowerDict(Quorum=42, Comment='')
        self.assertEqual(e.get('Quorum'), 42)

    def test_case(self):
        e = LowerDict(Quorum=42, Comment='')
        self.assertEqual(e['quorum'], 42)
        self.assertEqual(e['comment'], '')

    def test_contains(self):
        e = LowerDict(Annotation='a subtle yet profound statement', domain='society')
        self.assertTrue('annotation' in e.keys())
        self.assertTrue('Annotation ' in e.keys())
        self.assertTrue(' dOmAiN' in e.keys())
        self.assertFalse('SoCiEtY' in e.keys())
        self.assertFalse('annotat' in e.keys())

    def test_pop(self):
        e = LowerDict(Quorum=42, comment='no comment.')
        self.assertEqual(e.pop('quorum'), 42)
        self.assertIsNone(e.get('Quorum'))
        self.assertEqual(e.pop('COMMENT'), 'no comment.')
        self.assertIsNone(e.get('comment'))

    def test_update(self):
        e = LowerDict(comment='argh')
        self.assertEqual(e['Comment '], 'argh')
        d = {'Comment': 'no comment', 'Name': 'no name'}
        e.update(d)
        self.assertTrue('comment' in e)
        self.assertTrue('Comment ' in e)
        self.assertEqual(e['comment'], 'no comment')
        self.assertTrue('name' in e)
        self.assertTrue('Name' in e.keys())
        self.assertFalse('namee' in e.keys())

    def test_items(self):
        e = LowerDict(Name='James Bond', Number='007')
        d = {k: v for k, v in e.items()}
        self.assertEqual(d.pop('Number'), '007')


if __name__ == '__main__':
    unittest.main()



from .flowables import Flowable, FlowablesDict
import unittest
import json


flowables_json = '''{
"Flowables": [
    {
        "name": "carbon dioxide",
        "synonyms": [
            "124-38-9",
            "CO2"
        ]
    },
    {
        "name": "nitrous oxide",
        "synonyms": [
            "N2O",
            "dinitrogen monoxide",
            "10024972"
        ]
    },
    {
        "name": "methane",
        "synonyms": [
            "CH4",
            "methane (biogenic)",
            "methane, fossil"
        ]
    }
]
}'''


class FlowableTest(unittest.TestCase):
    def test_remove_cas_number(self):
        f = Flowable('methane', '74-82-8', 'CH4')
        self.assertIn('74-82-8', f)
        f.remove_term('000074828')
        self.assertNotIn('74-82-8', f)
        self.assertSetEqual(set(f.cas_numbers), set())

    def test_add_duplicate_cas(self):
        f = Flowable('methane', '74-82-8')
        self.assertSetEqual({t for t in f.terms}, {'methane', '74828', '74-82-8', '000074828', '000074-82-8'})
        c = 74828
        f.add_term(c)
        self.assertSetEqual({t for t in f.terms}, {'methane', '74828', '74-82-8', '000074828', '000074-82-8'})

    def test_rename(self):
        f = Flowable('CO2, biogenic', '124-38-9', 'carbon dioxide (biogenic)')
        self.assertEqual(f.name, 'CO2, biogenic')
        f.set_name('carbon dioxide (biogenic)')
        self.assertEqual(f.name, 'carbon dioxide (biogenic)')
        self.assertIn('CO2, biogenic', f)


class FlowablesTest(unittest.TestCase):
    def setUp(self):
        self.f = FlowablesDict()
        j = json.loads(flowables_json)
        for fb in j['Flowables']:
            self.f._add_from_dict(fb)

    def test_load_preserve_child(self):
        g = FlowablesDict()
        n = g.new_entry('nitrous oxide')
        o = g.new_entry('N2O', 'nitrous oxide', create_child=True)
        self.assertTrue(n.has_child(o))
        j = json.loads(flowables_json)
        for fb in j['Flowables']:
            g._add_from_dict(fb)
        self.assertEqual(g['N2O'], g['10024-97-2'])
        self.assertIs(g['N2O'], n)
        self.assertTrue(n.has_child(o))

    def test_cas_equivalence(self):
        self.assertEqual(self.f[124389], 'carbon dioxide')

    def test_add_set(self):
        self.f.new_entry('methane, biogenic', 'ch4', '74-82-8')
        self.assertEqual(self.f.get('000074-82-8'), 'methane')

    def test_return_type(self):
        self.assertIsInstance(self.f['carbon dioxide'], Flowable)

    def test_key_stability_under_rename(self):
        co2 = self.f['co2']
        self.assertEqual(co2.name, 'carbon dioxide')
        d = {co2: 420}
        self.f.set_name('co2')
        self.assertEqual(co2.name, 'CO2')
        self.assertEqual(d[self.f['co2']], 420)

    def test_serialize_deserialize(self):
        self.assertIn('124389', self.f)
        j = {self.f._entry_group: [f.serialize() for f in self.f._list_entries()]}
        self.assertEqual(len(j[self.f._entry_group]), len(self.f))
        co2 = next(f for f in j[self.f._entry_group] if f['name'] == 'carbon dioxide')
        self.assertListEqual(co2['synonyms'], ['CO2', '124-38-9'])
        g = FlowablesDict()
        for f in j[self.f._entry_group]:
            g._add_from_dict(f)
        self.assertEqual(len(g), len(self.f))
        self.assertIn('124389', g)


if __name__ == '__main__':
    unittest.main()



from ..synonym_dict import SynonymSet


class InvalidSubCompartment(Exception):
    pass


class NullContext(Exception):
    pass


class Compartment(SynonymSet):
    """
    A compartment is an environmental or social "compartment" that exchanges some "flow" with a technological "activity"
    in a process-flow product system.  Contexts are defined by a hierarchical structure and each instance has an
    optional 'parent' which is a proper superset, if present.

    If 'resources' or 'emissions' match any terms in a compartment, it is considered 'elementary', along with all its
    subcompartments.
    """
    @classmethod
    def null(cls):
        return cls('None')

    @property
    def is_null(self):
        return self._terms == {'None'}

    def add_term(self, term):
        if self.is_null:
            raise NullContext
        return super(Compartment, self).add_term(term)

    def add_child(self, other, force=False):
        if self.is_null:
            raise NullContext
        return super(Compartment, self).add_child(other, force=force)

    def __init__(self, *args, parent=None):
        super(Compartment, self).__init__(*args)
        self._parent = None
        self._subcompartments = set()
        if isinstance(parent, Compartment):
            self.parent = parent  # use setter
        elif parent is not None:
            raise TypeError('Parent must be a Compartment, not %s' % type(parent))

    @property
    def object(self):
        return self

    def top(self):
        if self.parent is None:
            return self
        return self.parent.top()

    def __iter__(self):
        if self.parent is not None:
            for k in self.parent:
                yield k
        yield str(self)

    def as_list(self):
        if self.parent is None:
            return [str(self)]
        else:
            return self.parent.as_list() + [str(self)]
        #return list(self)

    @property
    def base_terms(self):
        yield '; '.join(self.as_list())
        for x in super(Compartment, self).base_terms:
            yield x

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        if self._parent is not None:
            self._parent.deregister_subcompartment(self)
        self._parent = parent
        if parent is not None:
            parent.register_subcompartment(self)

    def is_subcompartment(self, comp):
        """
        for simplest semantics, a compartment is considered its own subcompartment. With None I could go either way.
        :param comp:
        :return:
        """
        if comp is self:
            return True
        if self._parent is not None:
            return self._parent.is_subcompartment(comp)
        return False

    @property
    def subcompartments(self):
        for s in self._subcompartments:
            yield s

    def register_subcompartment(self, sub):
        if self.is_null:
            raise NullContext
        if sub.parent is not self:
            raise InvalidSubCompartment('Parent %s: relationship does not exist: %s' % (self, sub))
        self._subcompartments.add(sub)

    def deregister_subcompartment(self, sub):
        if sub not in self._subcompartments:
            raise InvalidSubCompartment('Parent %s: subcompartment is not known: %s' % (self, sub))
        self._subcompartments.remove(sub)  # ??!

    @property
    def self_and_subcompartments(self):
        yield self
        for k in sorted(self._subcompartments, key=str):
            for t in k.self_and_subcompartments:
                yield t

    def serialize(self):
        d = super(Compartment, self).serialize()
        if self._parent is not None:
            d['parent'] = str(self.parent)
        return d



from setuptools import setup
import os

packages = []
root_dir = os.path.dirname(__file__)
if root_dir:
    os.chdir(root_dir)


# TODO
# Change `author`
# Change `author_email`
# Change `url`
# Change the PyPI classifiers: https://pypi.org/pypi?%3Aaction=list_classifiers


# Probably should be changed, __init__.py is no longer required for Python 3
for dirpath, dirnames, filenames in os.walk('synonym_dict'):
    # Ignore dirnames that start with '.'
    if '__init__.py' in filenames:
        pkg = dirpath.replace(os.path.sep, '.')
        if os.path.altsep:
            pkg = pkg.replace(os.path.altsep, '.')
        packages.append(pkg)


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


setup(
    name='synonym_dict',
    version="0.1.1",
    packages=packages,
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    license="BSD 3-clause",
    install_requires=[],
    url="https://github.com/bkuczenski/synonym_dict",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    description='A class that allows retrieval of a given object by any of its synonyms',
    classifiers=[
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
)
